==> audiochain-1.1.5-src.tgz: org/audiochain/devices/dynamics/EnvelopeAudioDataReader.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.dynamics;

import static java.lang.Math.*;

import java.io.IOException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;


public class EnvelopeAudioDataReader implements AudioDataReaderChainLink {

	public static enum EnvelopeDetectionType {
		Peak, RMS, RealLookAhead, Tangent
	}

	private EnvelopeDetectionType envelopeDetectionType = EnvelopeDetectionType.RealLookAhead;
	private AudioDataReader reader;
	private boolean leftEnvelopePassThrough;
	private boolean rightEnvelopePassThrough;
	private int[] envelope;
	private int leftMax;
	private int leftEnvelope;
	private int leftLast;
	private int rightMax;
	private int rightEnvelope;
	private int rightLast;

	private int[] leftRmsSampleBuffer = new int[11];
	private int rmsSampleBufferSize = leftRmsSampleBuffer.length;
	private int rmsSampleBufferSizeLessOne = rmsSampleBufferSize - 1;
	private int rmsSampleBufferSizeMidIndex = (rmsSampleBufferSize) / 2;
	private int[] rightRmsSampleBuffer = new int[rmsSampleBufferSize];

	// variables for Tangent and RealLookAhead
	private int leftDown;
	private int rightDown; // for peak detection when the curve goes downwards
	private float rightDecrement; // the distance between the max sample and its
									// following sample (when curve goes
									// downwards)
	private float leftDecrement;

	// variables for RealLookAhead
	private float leftDecrementMedian;
	private float leftCountMedian;
	private float rightDecrementMedian;
	private float rightCountMedian;

	public EnvelopeAudioDataReader(AudioDataReader reader) {
		this.reader = reader;
	}

	@Override
	public int read(int[] data) throws IOException {
		envelope = new int[data.length];

		final int len = reader.read(data);
		if (len == -1) {
			return len;
		}
		switch (envelopeDetectionType) {
		default:
		case Peak:
			peak(data, len);
			break;
		case RMS:
			rms(data, len);
			break;
		case Tangent:
			tangent(data, len);
			break;
		case RealLookAhead:
			realLookAhead(data, len);
			break;
		}

		return len;
	}

	private void realLookAhead(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			// left
			int l = 2 * i;
			int left = abs(data[l]);
			if (left >= leftLast && left >= leftEnvelope) {
				// upcoming
				leftMax = left;
				leftEnvelope = leftMax;
				leftDown = 0;
			} else {
				if (leftDown++ == 0) {
					leftDecrement = ((float) leftMax - (float) left);
					// get next peak value and sample count between this
					// peak and the next peak
					// decrement value is
					// "(this peak value - next peak value) / sample count"
					// (may be a positive or negative result)
					int vleftLast = left;
					int vleftMax = 0;
					boolean wasupcoming = false;
					for (int j = i; j < n; j++) {
						int nextPeak = abs(data[2 * j]);
						vleftMax = max(vleftMax, nextPeak);
						final float count = j - i;
						final float nextPeakVariance = nextPeak * 0.05F;
						final float nextPeakToCurrentPeakDiff = abs(nextPeak - left);
						if (nextPeak > vleftLast) {
							wasupcoming = true;
							// upcoming
						} else if (wasupcoming && nextPeakToCurrentPeakDiff <= nextPeakVariance && count > 2) {
							// downcoming
							leftDecrement = nextPeakToCurrentPeakDiff / count;
							break;
						} else if (count > leftCountMedian * 2) {
							leftDecrement = abs(vleftMax - left) / count;
							break;
						}
						leftCountMedian = (leftCountMedian + count) / 2;

						vleftLast = nextPeak;
					}
					// leftMax should be the previous sample before left at
					// this time+
					if (leftDecrement == 0 || leftDecrement > 2) {
						leftDecrement = max(min(1.9F, leftDecrementMedian), 0.9F);
					}
					leftDecrementMedian = (leftDecrementMedian + leftDecrement) / 2;
				}
				leftEnvelope -= leftDecrement;
			}
			int leftEnv = leftEnvelope;
			envelope[l] = leftEnv;

			if (leftEnvelopePassThrough) {
				data[l] = leftEnv;
			}

			leftLast = left;

			// right
			int r = l + 1;
			int right = abs(data[r]);
			if (right >= rightLast && right >= rightEnvelope) {
				// upcoming
				rightMax = right;
				rightEnvelope = rightMax;
				rightDown = 0;
			} else {
				if (rightDown++ == 0) {
					rightDecrement = ((float) rightMax - (float) right);
					// get next peak value and sample count between this
					// peak and the next peak
					// decrement value is
					// "(this peak value - next peak value) / sample count"
					// (may be a positive or negative result)
					int vrightLast = right;
					int vrightMax = 0;
					boolean wasupcoming = false;
					for (int j = i; j < n; j++) {
						int nextPeak = abs(data[2 * j + 1]);
						vrightMax = max(vrightMax, nextPeak);
						final float count = j - i;
						final float nextPeakVariance = nextPeak * 0.05F;
						final float nextPeakToCurrentPeakDiff = abs(nextPeak - right);
						if (nextPeak > vrightLast) {
							wasupcoming = true;
							// upcoming
						} else if (wasupcoming && nextPeakToCurrentPeakDiff <= nextPeakVariance && count > 2) {
							// downcoming
							rightDecrement = nextPeakToCurrentPeakDiff / count;
							break;
						} else if (count > rightCountMedian * 2) {
							rightDecrement = abs(vrightMax - right) / count;
							break;
						}
						rightCountMedian = (rightCountMedian + count) / 2;

						vrightLast = nextPeak;
					}
					// rightMax should be the previous sample before right
					// at
					// this time+
					if (rightDecrement == 0 || rightDecrement > 2) {
						rightDecrement = max(min(1.9F, rightDecrementMedian), 0.9F);
					}
					rightDecrementMedian = (rightDecrementMedian + rightDecrement) / 2;
				}
				rightEnvelope -= rightDecrement;
			}
			int rightEnv = rightEnvelope;
			envelope[r] = rightEnv;

			if (rightEnvelopePassThrough) {
				data[r] = rightEnv;
			}

			rightLast = right;
		}
	}

	private void tangent(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			// left
			int l = 2 * i;
			int left = abs(data[l]);
			if (left >= leftLast && left >= leftEnvelope) {
				// upcoming
				leftMax = left;
				leftEnvelope = leftMax;
				leftDown = 0;
			} else {
				if (leftDown++ == 0) {
					leftDecrement = ((float) leftMax - (float) left);
				}
				leftEnvelope -= leftDecrement;
			}

			int leftEnv = leftEnvelope;
			envelope[l] = leftEnv;

			if (leftEnvelopePassThrough) {
				data[l] = leftEnv;
			}

			leftLast = left;

			// right
			int r = l + 1;
			int right = abs(data[r]);
			if (right >= rightLast && right >= rightEnvelope) {
				// upcoming
				rightMax = right;
				rightEnvelope = rightMax;
				rightDown = 0;
			} else {
				if (rightDown++ == 0) {
					rightDecrement = ((float) rightMax - (float) right);
				}
				rightEnvelope -= rightDecrement;
			}

			int rightEnv = rightEnvelope;
			envelope[r] = rightEnv;

			if (rightEnvelopePassThrough) {
				data[r] = rightEnv;
			}

			rightLast = right;
		}
	}

	private void rms(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			// left
			System.arraycopy(leftRmsSampleBuffer, 1, leftRmsSampleBuffer, 0, rmsSampleBufferSizeLessOne);

			int idx = i;
			for (int j = rmsSampleBufferSizeMidIndex; j < rmsSampleBufferSize; j++, idx++) {
				if (idx < n) {
					leftRmsSampleBuffer[j] = data[2 * idx];
				}
			}

			double leftRms = 0;
			for (int j = 0; j < rmsSampleBufferSize; j++) {
				int rmsSample = leftRmsSampleBuffer[j];
				leftRms += rmsSample * rmsSample;
			}
			leftRms /= rmsSampleBufferSize;
			leftRms = sqrt(leftRms);

			int l = 2 * i;
			envelope[l] = (int) leftRms;

			if (leftEnvelopePassThrough) {
				data[l] = (int) leftRms;
			}

			// right
			System.arraycopy(rightRmsSampleBuffer, 1, rightRmsSampleBuffer, 0, rmsSampleBufferSizeLessOne);

			idx = i;
			for (int j = rmsSampleBufferSizeMidIndex; j < rmsSampleBufferSize; j++, idx++) {
				if (idx < n) {
					rightRmsSampleBuffer[j] = data[2 * idx + 1];
				}
			}

			double rightRms = 0;
			for (int j = 0; j < rmsSampleBufferSize; j++) {
				int rmsSample = rightRmsSampleBuffer[j];
				rightRms += rmsSample * rmsSample;
			}
			rightRms /= rmsSampleBufferSize;
			rightRms = sqrt(rightRms);

			int r = l + 1;
			envelope[r] = (int) rightRms;

			if (rightEnvelopePassThrough) {
				data[r] = (int) rightRms;
			}
		}
	}

	private void peak(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			int l = 2 * i;
			int left = abs(data[l]);
			if (left > leftLast) {
				// upcoming
				leftMax = max(left, leftMax);
			} else if (leftMax != 0) {
				leftEnvelope = leftMax;
				leftMax = 0;
			}

			int leftEnv = max(left, leftEnvelope);
			envelope[l] = leftEnv;

			if (leftEnvelopePassThrough) {
				data[l] = leftEnv;
			}
			leftLast = left;

			int r = l + 1;
			int right = abs(data[r]);
			if (right > rightLast) {
				// upcoming
				rightMax = max(right, rightMax);
			} else if (rightMax != 0) {
				rightEnvelope = rightMax;
				rightMax = 0;
			}

			int rightEnv = max(right, rightEnvelope);
			envelope[r] = rightEnv;

			if (rightEnvelopePassThrough) {
				data[r] = rightEnv;
			}
			rightLast = right;
		}
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}

	@Override
	public void stop() {
		reader.stop();
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public int[] getEnvelope() {
		return envelope;
	}

	public boolean isLeftEnvelopePassThrough() {
		return leftEnvelopePassThrough;
	}

	public void setLeftEnvelopePassThrough(boolean leftEnvelopePassThrough) {
		this.leftEnvelopePassThrough = leftEnvelopePassThrough;
	}

	public boolean isRightEnvelopePassThrough() {
		return rightEnvelopePassThrough;
	}

	public void setRightEnvelopePassThrough(boolean rightEnvelopePassThrough) {
		this.rightEnvelopePassThrough = rightEnvelopePassThrough;
	}

	public EnvelopeDetectionType getEnvelopeDetectionType() {
		return envelopeDetectionType;
	}

	public void setEnvelopeDetectionType(EnvelopeDetectionType envelopeDetectionType) {
		this.envelopeDetectionType = envelopeDetectionType;
	}

}

==> audiochain-1.1.6-src.tgz: org/audiochain/devices/dynamics/EnvelopeAudioDataReader.java <==
/**
   AudioChain Copyright (C) 2012-2013 Florian Paul (www.audiochain.org)
   
   This program is free software; you can redistribute it and/or modify it under the terms of the 
   GNU General Public License as published by the Free Software Foundation; either version 3 of the 
   License, or (at your option) any later version.
   
   This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without 
   even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General 
   Public License for more details.
   
   You should have received a copy of the GNU General Public License along with this program; 
   if not, see <http://www.gnu.org/licenses/>. 
 */
package org.audiochain.devices.dynamics;

import static java.lang.Math.abs;
import static java.lang.Math.max;
import static java.lang.Math.min;
import static java.lang.Math.sqrt;

import java.io.IOException;

import org.audiochain.io.AudioDataReader;
import org.audiochain.io.AudioDataReaderChainLink;


public class EnvelopeAudioDataReader implements AudioDataReaderChainLink {

	public static enum EnvelopeDetectionType {
		Peak, RMS, RealLookAhead, Tangent
	}

	private EnvelopeDetectionType envelopeDetectionType = EnvelopeDetectionType.RealLookAhead;
	private AudioDataReader reader;
	private boolean leftEnvelopePassThrough;
	private boolean rightEnvelopePassThrough;
	private int[] envelope;
	private int leftMax;
	private int leftEnvelope;
	private int leftLast;
	private int rightMax;
	private int rightEnvelope;
	private int rightLast;

	private int[] leftRmsSampleBuffer = new int[11];
	private int rmsSampleBufferSize = leftRmsSampleBuffer.length;
	private int rmsSampleBufferSizeLessOne = rmsSampleBufferSize - 1;
	private int rmsSampleBufferSizeMidIndex = (rmsSampleBufferSize) / 2;
	private int[] rightRmsSampleBuffer = new int[rmsSampleBufferSize];

	// variables for Tangent and RealLookAhead
	private int leftDown;
	private int rightDown; // for peak detection when the curve goes downwards
	private float rightDecrement; // the distance between the max sample and its
									// following sample (when curve goes
									// downwards)
	private float leftDecrement;

	// variables for RealLookAhead
	private float leftDecrementMedian;
	private float leftCountMedian;
	private float rightDecrementMedian;
	private float rightCountMedian;

	public EnvelopeAudioDataReader(AudioDataReader reader) {
		this.reader = reader;
	}

	@Override
	public int read(int[] data) throws IOException {
		envelope = new int[data.length];

		final int len = reader.read(data);
		if (len == -1) {
			return len;
		}
		switch (envelopeDetectionType) {
		default:
		case Peak:
			peak(data, len);
			break;
		case RMS:
			rms(data, len);
			break;
		case Tangent:
			tangent(data, len);
			break;
		case RealLookAhead:
			realLookAhead(data, len);
			break;
		}

		return len;
	}

	private void realLookAhead(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			// left
			int l = 2 * i;
			int left = abs(data[l]);
			if (left >= leftLast && left >= leftEnvelope) {
				// upcoming
				leftMax = left;
				leftEnvelope = leftMax;
				leftDown = 0;
			} else {
				if (leftDown++ == 0) {
					leftDecrement = ((float) leftMax - (float) left);
					// get next peak value and sample count between this
					// peak and the next peak
					// decrement value is
					// "(this peak value - next peak value) / sample count"
					// (may be a positive or negative result)
					int vleftLast = left;
					int vleftMax = 0;
					boolean wasupcoming = false;
					for (int j = i; j < n; j++) {
						int nextPeak = abs(data[2 * j]);
						vleftMax = max(vleftMax, nextPeak);
						final float count = j - i;
						final float nextPeakVariance = nextPeak * 0.05F;
						final float nextPeakToCurrentPeakDiff = abs(nextPeak - left);
						if (nextPeak > vleftLast) {
							wasupcoming = true;
							// upcoming
						} else if (wasupcoming && nextPeakToCurrentPeakDiff <= nextPeakVariance && count > 2) {
							// downcoming
							leftDecrement = nextPeakToCurrentPeakDiff / count;
							break;
						} else if (count > leftCountMedian * 2) {
							leftDecrement = abs(vleftMax - left) / count;
							break;
						}
						leftCountMedian = (leftCountMedian + count) / 2;

						vleftLast = nextPeak;
					}
					// leftMax should be the previous sample before left at
					// this time+
					if (leftDecrement == 0 || leftDecrement > 2) {
						leftDecrement = max(min(1.9F, leftDecrementMedian), 0.9F);
					}
					leftDecrementMedian = (leftDecrementMedian + leftDecrement) / 2;
				}
				leftEnvelope -= leftDecrement;
			}
			int leftEnv = leftEnvelope;
			envelope[l] = leftEnv;

			if (leftEnvelopePassThrough) {
				data[l] = leftEnv;
			}

			leftLast = left;

			// right
			int r = l + 1;
			int right = abs(data[r]);
			if (right >= rightLast && right >= rightEnvelope) {
				// upcoming
				rightMax = right;
				rightEnvelope = rightMax;
				rightDown = 0;
			} else {
				if (rightDown++ == 0) {
					rightDecrement = ((float) rightMax - (float) right);
					// get next peak value and sample count between this
					// peak and the next peak
					// decrement value is
					// "(this peak value - next peak value) / sample count"
					// (may be a positive or negative result)
					int vrightLast = right;
					int vrightMax = 0;
					boolean wasupcoming = false;
					for (int j = i; j < n; j++) {
						int nextPeak = abs(data[2 * j + 1]);
						vrightMax = max(vrightMax, nextPeak);
						final float count = j - i;
						final float nextPeakVariance = nextPeak * 0.05F;
						final float nextPeakToCurrentPeakDiff = abs(nextPeak - right);
						if (nextPeak > vrightLast) {
							wasupcoming = true;
							// upcoming
						} else if (wasupcoming && nextPeakToCurrentPeakDiff <= nextPeakVariance && count > 2) {
							// downcoming
							rightDecrement = nextPeakToCurrentPeakDiff / count;
							break;
						} else if (count > rightCountMedian * 2) {
							rightDecrement = abs(vrightMax - right) / count;
							break;
						}
						rightCountMedian = (rightCountMedian + count) / 2;

						vrightLast = nextPeak;
					}
					// rightMax should be the previous sample before right
					// at
					// this time+
					if (rightDecrement == 0 || rightDecrement > 2) {
						rightDecrement = max(min(1.9F, rightDecrementMedian), 0.9F);
					}
					rightDecrementMedian = (rightDecrementMedian + rightDecrement) / 2;
				}
				rightEnvelope -= rightDecrement;
			}
			int rightEnv = rightEnvelope;
			envelope[r] = rightEnv;

			if (rightEnvelopePassThrough) {
				data[r] = rightEnv;
			}

			rightLast = right;
		}
	}

	private void tangent(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			// left
			int l = 2 * i;
			int left = abs(data[l]);
			if (left >= leftLast && left >= leftEnvelope) {
				// upcoming
				leftMax = left;
				leftEnvelope = leftMax;
				leftDown = 0;
			} else {
				if (leftDown++ == 0) {
					leftDecrement = ((float) leftMax - (float) left);
				}
				leftEnvelope -= leftDecrement;
			}

			int leftEnv = leftEnvelope;
			envelope[l] = leftEnv;

			if (leftEnvelopePassThrough) {
				data[l] = leftEnv;
			}

			leftLast = left;

			// right
			int r = l + 1;
			int right = abs(data[r]);
			if (right >= rightLast && right >= rightEnvelope) {
				// upcoming
				rightMax = right;
				rightEnvelope = rightMax;
				rightDown = 0;
			} else {
				if (rightDown++ == 0) {
					rightDecrement = ((float) rightMax - (float) right);
				}
				rightEnvelope -= rightDecrement;
			}

			int rightEnv = rightEnvelope;
			envelope[r] = rightEnv;

			if (rightEnvelopePassThrough) {
				data[r] = rightEnv;
			}

			rightLast = right;
		}
	}

	private void rms(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			// left
			System.arraycopy(leftRmsSampleBuffer, 1, leftRmsSampleBuffer, 0, rmsSampleBufferSizeLessOne);

			int idx = i;
			for (int j = rmsSampleBufferSizeMidIndex; j < rmsSampleBufferSize; j++, idx++) {
				if (idx < n) {
					leftRmsSampleBuffer[j] = data[2 * idx];
				}
			}

			double leftRms = 0;
			for (int j = 0; j < rmsSampleBufferSize; j++) {
				int rmsSample = leftRmsSampleBuffer[j];
				leftRms += rmsSample * rmsSample;
			}
			leftRms /= rmsSampleBufferSize;
			leftRms = sqrt(leftRms);

			int l = 2 * i;
			envelope[l] = (int) leftRms;

			if (leftEnvelopePassThrough) {
				data[l] = (int) leftRms;
			}

			// right
			System.arraycopy(rightRmsSampleBuffer, 1, rightRmsSampleBuffer, 0, rmsSampleBufferSizeLessOne);

			idx = i;
			for (int j = rmsSampleBufferSizeMidIndex; j < rmsSampleBufferSize; j++, idx++) {
				if (idx < n) {
					rightRmsSampleBuffer[j] = data[2 * idx + 1];
				}
			}

			double rightRms = 0;
			for (int j = 0; j < rmsSampleBufferSize; j++) {
				int rmsSample = rightRmsSampleBuffer[j];
				rightRms += rmsSample * rmsSample;
			}
			rightRms /= rmsSampleBufferSize;
			rightRms = sqrt(rightRms);

			int r = l + 1;
			envelope[r] = (int) rightRms;

			if (rightEnvelopePassThrough) {
				data[r] = (int) rightRms;
			}
		}
	}

	private void peak(int[] data, final int len) {
		for (int i = 0, n = len / 2; i < n; i++) {
			int l = 2 * i;
			int left = abs(data[l]);
			if (left > leftLast) {
				// upcoming
				leftMax = max(left, leftMax);
			} else if (leftMax != 0) {
				leftEnvelope = leftMax;
				leftMax = 0;
			}

			int leftEnv = max(left, leftEnvelope);
			envelope[l] = leftEnv;

			if (leftEnvelopePassThrough) {
				data[l] = leftEnv;
			}
			leftLast = left;

			int r = l + 1;
			int right = abs(data[r]);
			if (right > rightLast) {
				// upcoming
				rightMax = max(right, rightMax);
			} else if (rightMax != 0) {
				rightEnvelope = rightMax;
				rightMax = 0;
			}

			int rightEnv = max(right, rightEnvelope);
			envelope[r] = rightEnv;

			if (rightEnvelopePassThrough) {
				data[r] = rightEnv;
			}
			rightLast = right;
		}
	}

	@Override
	public void close() throws IOException {
		reader.close();
	}

	@Override
	public void stop() {
		reader.stop();
	}

	@Override
	public void seek(long framePosition) throws IOException {
		reader.seek(framePosition);
	}

	@Override
	public void setSource(AudioDataReader source) {
		this.reader = source;
	}

	@Override
	public AudioDataReader getSource() {
		return reader;
	}

	public int[] getEnvelope() {
		return envelope;
	}

	public boolean isLeftEnvelopePassThrough() {
		return leftEnvelopePassThrough;
	}

	public void setLeftEnvelopePassThrough(boolean leftEnvelopePassThrough) {
		this.leftEnvelopePassThrough = leftEnvelopePassThrough;
	}

	public boolean isRightEnvelopePassThrough() {
		return rightEnvelopePassThrough;
	}

	public void setRightEnvelopePassThrough(boolean rightEnvelopePassThrough) {
		this.rightEnvelopePassThrough = rightEnvelopePassThrough;
	}

	public EnvelopeDetectionType getEnvelopeDetectionType() {
		return envelopeDetectionType;
	}

	public void setEnvelopeDetectionType(EnvelopeDetectionType envelopeDetectionType) {
		this.envelopeDetectionType = envelopeDetectionType;
	}

}
